# Copyright (C) 2015 zulily, llc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""helper - a module containing a collection useful object manipulations"""

import json
import re
import yaml

def object_filter(objects, grep):
    """Filter out any objects that do not have attributes with values matching
    *all* regular expressions present in grep (AND, essentially)

    :param objects ADObject: A list of ADObjects
    :param grep list: A list of regular expressions that must match for filtering

    :return: A list of filtered ADObjects
    :rtype: list
    """
    filtered = []
    if grep:
        for ad_object in objects:
            o_string = ' '.join([value for value in ad_object.to_dict().values()
                                 if isinstance(value, str)])
            skip = False
            for regex in grep:
                if not re.search(regex, o_string, re.M|re.S|re.I):
                    skip = True
                    break
            if not skip:
                filtered.append(ad_object)

        return filtered
    else:
        return objects


def serialize(ad_objects, output_format='json', indent=2, attributes_only=False):
    """Serialize the object to the specified format

    :param ad_objects list: A list of ADObjects to serialize
    :param output_format str: The output format, json or yaml.  Defaults to json
    :param indent int: The number of spaces to indent, defaults to 2
    :param attributes only: Only serialize the attributes found in the first record of the list
        of ADObjects

    :return: A serialized, formatted representation of the list of ADObjects
    :rtype: str
    """

    # If the request is to only show attributes for objects returned
    # in the query, overwrite ad_objects with only those attributes present in
    # the first object in the list
    if attributes_only:
        ad_objects = [key for key in sorted(ad_objects[0].keys())]

    if output_format == 'json':
        return json.dumps(ad_objects, indent=indent, ensure_ascii=False, sort_keys=True)
    elif output_format == 'yaml':
        return yaml.dump(sorted(ad_objects), indent=indent)






# Copyright (C) 2015 zulily, llc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ad_group"""

import logging

from pudl.ad_object import ADObject

class ADGroup(ADObject):
    """A class to represent AD group objects.  Includes a number of
    helper methods, particularly object-factory related.

    ADGroup objects have minimal depth, with attributes set to
    strings or lists.  Available attributes are dependent
    on the results returned by the LDAP query.

    In its current implementation, the memberOf attribute
    is not expanded.  The member attribute is however flattened out.

    """


    def group(self, base_dn, samaccountname, attributes=(), explicit_membership_only=False):
        """Produces a single, populated ADGroup object through the object factory.
        Does not populate attributes for the caller instance.

        sAMAccountName may not be present in group objects in modern AD schemas.
        Searching by common name and object class (group) may be an alternative
        approach if required in the future.

        :param str base_dn: The base DN to search within
        :param str samaccountname: The group's sAMAccountName
        :param list attributes: Object attributes to populate, defaults to all

        :return: A populated ADGroup object
        :rtype: ADGroup
        """

        groups = self.groups(base_dn, samaccountnames=[samaccountname], attributes=attributes,
                             explicit_membership_only=explicit_membership_only)

        try:
            # Usually we will find a match, but perhaps not always
            return groups[0]
        except IndexError:
            logging.info("%s - unable to retrieve object from AD by sAMAccountName", samaccountname)


    def groups(self, base_dn, samaccountnames=(), attributes=(), explicit_membership_only=False):
        """Gathers a list of ADGroup objects

        sAMAccountName may not be present in group objects in modern AD schemas.
        Searching by common name and object class (group) may be an alternative
        approach if required in the future.

        :param str base_dn: The base DN to search within
        :param list samaccountnames: A list of group names for which objects will be
            created, defaults to all groups if unspecified
        :param list attributes: Object attributes to populate, defaults to all

        :return: A list of populated ADGroup objects
        :rtype: list
        """
        ad_groups = []

        search_filter = '(&(objectClass=group)(!(objectClass=user))(!(objectClass=computer)){0})'
        # If no samaccountnames specified, filter will pull all group objects under
        # base_dn
        if not samaccountnames:
            search_filter = search_filter.format('(sAMAccountName=*)')
        else:
            if len(samaccountnames) == 1:
                group_names = '(sAMAccountName={0})'.format(samaccountnames[0])
            else:
                group_names = '(|{0})'.format(''.join(['(sAMAccountName={0})'.\
                                                       format(group) for group
                                                       in samaccountnames]))

            search_filter = search_filter.format(group_names)


        logging.debug('%s Search filter: %s', self.__class__.__name__, search_filter)

        results = self.adq.search(base_dn, search_filter, attributes)

        for search_result in results:
            adg = self._object_factory(search_result)
            if not explicit_membership_only and 'member' in dir(adg):
                member = [u[0] for u in
                          self.adq.search(base_dn, '(memberOf:1.2.840.113556.1.4.1941:={0})'.\
                          format(search_result[0]), attributes=['member'])]
                adg.member = member
            ad_groups.append(adg)

        return ad_groups




# Copyright (C) 2015 zulily, llc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ad_query"""

import logging
import ldap

# Set a default LDAP URL, port 389 as we are using LDAP+TLS and not LDAPS
LDAP_URL = 'ldap://ldap:389'

# LDAP options related to trusting the identity of the remote server
#pylint: disable=no-member
LDAP_OPTIONS = ((ldap.OPT_PROTOCOL_VERSION, ldap.VERSION3),
                (ldap.OPT_REFERRALS, 0))

LDAP_OPTIONS_TLS_NO_VERIFY = ((ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_ALLOW),
                              (ldap.OPT_PROTOCOL_VERSION, ldap.VERSION3),
                              (ldap.OPT_REFERRALS, 0))
#pylint: enable=no-member

# Limit the number of results per page, with a default that should generally
# be below the size of the limit imposed by most AD servers
PAGE_SIZE = 300

# Require identity trust is established with the server by default
TLS_NO_VERIFY = False

class ADQuery(object):  #pylint: disable=too-few-public-methods
    """Query Active directory with python-ldap.  May be used directly, but is most
    commonly used indirectly via ADObject-based classes.  All connections
    require TLS.
    """

    def __init__(self, user, password,
                 ldap_url=LDAP_URL,
                 tls_no_verify=TLS_NO_VERIFY,
                 page_size=PAGE_SIZE):
        """The ADQuery constructor

        :param str user: The LDAP user to connect as
        :param str password: The LDAP user's password
        :param str ldap_url: The url, defaults to *{0}*
        :param bool tls_no_verify: If True, connect to servers with certificates not signed
            by an authority we trust, defaults to False
        :param int page_size: The max result set size, per page, defaults to *{1}*
        """.format(LDAP_URL, PAGE_SIZE)
        if tls_no_verify:
            ldap_options = LDAP_OPTIONS_TLS_NO_VERIFY
        else:
            ldap_options = LDAP_OPTIONS

        # Setup logging, assumes a root logger already exists with handlers
        self.logger = logging.getLogger(__name__)

        # Set LDAP options for the connection
        for setting in ldap_options:
            ldap.set_option(setting[0], setting[1])

        self.ldap = ldap.initialize(ldap_url)
        self.sprc = ldap.controls.SimplePagedResultsControl(True, page_size, '')
        self.user = user
        self.password = password
        self.page_size = page_size

        # Open the connection
        self._open()


    def search(self, base_dn, search_filter, attributes=()):
        """Perform an AD search

        :param str base_dn: The base DN to search within
        :param str search_filter: The search filter to apply, such as:
          *objectClass=person*
        :param list attributes: Object attributes to populate, defaults to all
        """
        results = []
        page = 0
        while page == 0 or self.sprc.cookie:
            page += 1
            #pylint: disable=no-member
            message_id = self.ldap.search_ext(base_dn, ldap.SCOPE_SUBTREE,
                                              search_filter, attributes,
                                              serverctrls=[self.sprc])
            #pylint: enable=no-member
            data, server_controls = self.ldap.result3(message_id)[1::2]
            self.sprc.cookie = server_controls[0].cookie
            logging.debug('%s - Page %s results: %s',  \
                          self.__class__.__name__, page, ', '.join(k[0] for k in data))
            results += [u for u in data]

        return results


    def _open(self):
        """Bind, use tls"""
        try:
            self.ldap.start_tls_s()
        #pylint: disable=no-member
        except ldap.CONNECT_ERROR:
        #pylint: enable=no-member
            logging.error('Unable to establish a connection to the LDAP server, ' + \
                          'please check the connection string ' + \
                          'and ensure the remote certificate is signed by a trusted authority.')
            raise

        self.ldap.simple_bind_s(self.user, self.password)




#! /usr/bin/env python
#
# Copyright (C) 2015 zulily, llc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""pudl cli"""

from __future__ import print_function

import argparse
import getpass
import logging
import os
import sys

from pudl import __version__ as pudl_version
from pudl.ad_computer import ADComputer
from pudl.ad_group import ADGroup
from pudl.ad_query import ADQuery
from pudl.ad_user import ADUser
from pudl.helper import object_filter, serialize


def main():
    """Do some stuff"""
    # Parse all command line argument
    args = parse_arguments().parse_args()

    # Setup logging
    configure_logging(args)
    logging.debug(args)

    # Prompt for a password if necessary
    if not args.password:
        password = getpass.getpass(prompt='Password ({0}): '.format(args.user))
    else:
        password = args.password

    # Create an instance of ADQuery which sets up a single
    # connection used for querying, for all AD object types
    ldap_url = 'ldap://{0}:{1}'.format(args.host, args.port)
    adq = ADQuery(user=args.user, password=password, page_size=args.page_size,
                  ldap_url=ldap_url, tls_no_verify=args.tls_no_verify)
    if args.subcommand == 'user':
        adu = ADUser(adq)
        users = adu.users(base_dn=args.base_dn, attributes=args.attributes,
                          samaccountnames=args.samaccountnames,
                          explicit_membership_only=args.explicit_membership_only)
        users = object_filter(users, args.grep)
        print(serialize([u.to_dict() for u in users],
                        output_format=args.output_format, attributes_only=args.attributes_only))
    elif args.subcommand == 'group':
        adg = ADGroup(adq)
        groups = adg.groups(base_dn=args.base_dn, attributes=args.attributes,
                            samaccountnames=args.samaccountnames,
                            explicit_membership_only=args.explicit_membership_only)
        groups = object_filter(groups, args.grep)
        print(serialize([g.to_dict() for g in groups],
                        output_format=args.output_format, attributes_only=args.attributes_only))
    elif args.subcommand == 'computer':
        adg = ADComputer(adq)
        computers = adg.computers(base_dn=args.base_dn, attributes=args.attributes,
                                  samaccountnames=args.samaccountnames)
        computers = object_filter(computers, args.grep)
        print(serialize([c.to_dict() for c in computers],
                        output_format=args.output_format, attributes_only=args.attributes_only))


def configure_logging(args):
    """Logging to console"""
    log_format = logging.Formatter('%(levelname)s:%(name)s:line %(lineno)s:%(message)s')
    log_level = logging.INFO if args.verbose else logging.WARN
    log_level = logging.DEBUG if args.debug else log_level
    console = logging.StreamHandler()
    console.setFormatter(log_format)
    console.setLevel(log_level)
    root_logger = logging.getLogger()
    if len(root_logger.handlers) == 0:
        root_logger.addHandler(console)
    root_logger.setLevel(log_level)
    root_logger.handlers[0].setFormatter(log_format)
    logging.getLogger(__name__)


def parse_arguments():
    """Collect command-line arguments.  Let the caller run parse_args(), as
    sphinx-argparse requires a function that returns an instance of
    argparse.ArgumentParser
    """
    # Pull a few settings from the environment, should they exist
    base_dn = os.environ['PUDL_BASE_DN'] if 'PUDL_BASE_DN' in os.environ \
              else 'OU=Departments,DC=example,DC=com'
    domain = os.environ['PUDL_DOMAIN'].upper() if 'PUDL_DOMAIN' in os.environ else 'EXAMPLE'
    page_size = os.environ['PUDL_PAGE_SIZE'].upper() if 'PUDL_PAGE_SIZE' in os.environ else 300
    tls_no_verify = bool(os.environ['PUDL_TLS_NO_VERIFY'].lower().capitalize()) \
                    if 'PUDL_TLS_NO_VERIFY' in os.environ else False

    parser = argparse.ArgumentParser(prog='pudl',
                                     description='A script for interacting with Active ' + \
                                     'Directory, leveraging python-ldap')
    parser.add_argument('-V', '--version', action='version', version='pudl v' + pudl_version,
                        help="Print the version number and exit")
    subparsers = parser.add_subparsers(dest='subcommand', help='Sub-command help')
    parser_common = subparsers.add_parser('common', add_help=False)
    parser_common.add_argument('--user', '-u', action='store', dest='user',
                               help='The ldap user (bind dn) to connect as. ' + \
                               'The full DN will work, or often, just the CN may be ' + \
                               'sufficient, such as "John Smith", or more commonly, ' + \
                               'specify the domain and sAMAccountName. Defaults to ' + \
                               '{0}\\username.  The domain '.format(domain) + \
                               'portion may be overridden with PUDL_DOMAIN',
                               default='{0}\\{1}'.format(domain, getpass.getuser()))
    parser_common.add_argument('--password', '-p', action='store',
                               dest='password', help="The connecting user's password")
    parser_common.add_argument('--host', '-H', action='store',
                               dest='host', help='The AD/LDAP host, defaults to ldap',
                               default='ldap')
    parser_common.add_argument('--port', '-P', action='store', dest='port',
                               help='The ldap port, defaults to 389.  389 is ' + \
                               'is the standard port', type=int, default=389)
    parser_common.add_argument('--page-size', '-s', action='store', dest='page_size',
                               help='The ldap results are paged, specify the ' + \
                               'number of results per page, defaults to ' + \
                               '{0}. May be overridden with PUDL_PAGE_SIZE'.format(page_size),
                               type=int, default=page_size)
    parser_common.add_argument('--base-dn', '-b', action='store',
                               dest='base_dn', default=base_dn,
                               help="The Base DN to use, defaults to {0}. ".format(base_dn) + \
                               "May be overridden with PUDL_BASE_DN")
    parser_common.add_argument('--attribute', '-a', action='append',
                               dest='attributes', metavar='ATTRIBUTE',
                               help="Attributes to include in results objects.  Note that " + \
                               "any nested objects return all attributes.  Maybe be used " + \
                               "multiple times, and if not specified, all " + \
                               "attributes are included in top-level objects")
    parser_common.add_argument('--grep', '-g', action='append', dest='grep',
                               help='Filter results to only those matching the specified ' + \
                               'regular expression (compares against all attributes). ' + \
                               'May be used multiple times')
    parser_common.add_argument('--attributes-only', '-A', action='store_true',
                               dest='attributes_only', help="Only display a list of attributes " + \
                               "that are present for the object type returned by the LDAP query")
    parser_common.add_argument('--output-format', '-f', action='store', dest='output_format',
                               choices=['json', 'yaml'], default='json',
                               help="Output format, defaults to json.")
    parser_common.add_argument('--verbose', '-v', action='store_true', dest='verbose',
                               help='Turn on verbose output', default=False)
    parser_common.add_argument('--debug', '-d', action='store_true', dest='debug', default=False,
                               help="Print out debugging information, very chatty")
    parser_common.add_argument('--tls-no-verify', '-V', action='store_true',
                               dest='tls_no_verify',
                               default=tls_no_verify, help="Don't verify the authenticity " + \
                               "of the server's certificate, defaults to " + \
                               "{0} and may be overridden with ".format(tls_no_verify) + \
                               "PUDL_TLS_NO_VERIFY")
    parser_user = subparsers.add_parser('user', parents=[parser_common], conflict_handler='resolve',
                                        help='Pull user objects from AD')
    parser_user.add_argument(nargs="*", dest='samaccountnames',
                             help='sAMAccountNames for any user objects that are to be ' + \
                             'looked up. If unspecified, returns all users under the base ' + \
                             'DN provided')
    parser_user.add_argument('--explicit-membership-only', '-e', action='store_true',
                             dest='explicit_membership_only', default=False,
                             help="Only show membership for users that is explicit, " + \
                             "not taking into account group nesting.  Defaults to False")
    parser_group = subparsers.add_parser('group', parents=[parser_common],
                                         conflict_handler='resolve',
                                         help='Pull group objects from AD')
    parser_group.add_argument(nargs="*", dest='samaccountnames',
                              help="sAMAccountNames for any group objects that are to be " + \
                              'looked up. If unspecified, returns all groups under the base ' + \
                              'DN provided. sAMAccountName may not be present in group ' + \
                              'objects in modern AD schemas')
    parser_group.add_argument('--explicit-membership-only', '-e', action='store_true',
                              dest='explicit_membership_only', default=False,
                              help="Only show membership for users that is explicit, " + \
                              "not taking into account group nesting.  Defaults to False")
    parser_computer = subparsers.add_parser('computer', parents=[parser_common],
                                            conflict_handler='resolve',
                                            help='Pull computer objects from AD')
    parser_computer.add_argument(nargs="*", dest='samaccountnames',
                                 help="sAMAccountNames for any computer objects that are to be " + \
                                 'looked up. If unspecified, returns all computers under ' + \
                                 'the base DN provided.')

    # sphinx is not add_help=False aware...
    del subparsers.choices['common']

    return parser


if __name__ == '__main__':
    sys.exit(main())



# Copyright (C) 2015 zulily, llc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ad_user"""

import logging

from pudl.ad_object import ADObject

class ADUser(ADObject):
    """A class to represent AD user objects.  Includes a number of
    helper methods, particularly object-factory related.

    ADUser objects have minimal depth, with attributes set to
    strings or lists.  Available attributes are dependent
    on the results returned by the LDAP query.
    """

    # Some refactoring may be considered in the future that would
    # involve passing the sAMAccountName to a contstructor override,
    # and possibly moving users() to become static.  Otherwise,
    # instead of user() creating and returning a single new object,
    # perhaps just populate the current ADUser instance, which
    # could make a little more sense



    def user(self, base_dn, samaccountname, attributes=(), explicit_membership_only=False):
        """Produces a single, populated ADUser object through the object factory.
        Does not populate attributes for the caller instance.

        :param str base_dn: The base DN to search within
        :param str samaccountname: The user's sAMAccountName
        :param list attributes: Object attributes to populate, defaults to all
        :param bool explicit_membership_only: If set True, memberof will only
            list groups for which the user is a directly referenced member

        :return: A populated ADUser object
        :rtype: ADUser
        """

        users = self.users(base_dn, samaccountnames=[samaccountname],
                           attributes=attributes, explicit_membership_only=explicit_membership_only)

        try:
            # Usually we will find a match, but perhaps not always
            return users[0]
        except IndexError:
            logging.info("%s - unable to retrieve object from AD by sAMAccountName", samaccountname)



    def users(self, base_dn, samaccountnames=(), attributes=(), explicit_membership_only=False):
        """Gathers a list of ADUser objects

        :param str base_dn: The base DN to search within
        :param list attributes: Object attributes to populate, defaults to all
        :param list samaccountnames: A list of usernames for which objects will be
            created, defaults to all users if unspecified
        :param bool explicit_membership_only: If set True, memberof will only
            list groups for which users are directly referenced members

        :return: A list of populated ADUser objects
        :rtype: list
        """
        ad_users = []

        search_filter = '(&(objectClass=user)(!(objectClass=group))(!(objectClass=computer)){0})'
        # If no samaccountnames specified, filter will pull all user objects under
        # base_dn
        if not samaccountnames:
            search_filter = search_filter.format('(sAMAccountName=*)')
        else:
            # Extensible filter: http://bit.ly/1Qh4eyV
            if len(samaccountnames) == 1:
                account_names = '(sAMAccountName={0})'.format(samaccountnames[0])
            else:
                account_names = '(|{0})'.format(''.join(['(sAMAccountName={0})'.format(username) \
                                                for username in samaccountnames]))

            search_filter = search_filter.format(account_names)


        logging.debug('%s Search filter: %s', self.__class__.__name__, search_filter)

        results = self.adq.search(base_dn, search_filter, attributes)

        for search_result in results:
            adu = self._object_factory(search_result)
            # Each results index 0 of the tuple is the DN
            if not explicit_membership_only and 'memberof' in dir(adu):
                memberof = [g[0] for g in self.adq.search(base_dn,
                                                          '(member:1.2.840.113556.1.4.1941:={0})'.\
                                                          format(search_result[0]),
                                                          attributes=['memberof'])]
                adu.memberof = memberof
            ad_users.append(adu)


        return ad_users


    def is_member(self, group_distinguishedname):
        """For the current ADUser instance, determine if
        the user is a member of a specific group (the group DN is used).
        The result may not be accurate if explicit_membership_only was set to
        True when the object factory method (user() or users()) was
        called.

        :param str group_distinguishedname: The group DistinguishedName

        :return: A boolean indicating whether or not the user is a member of the group
        :rtype: bool
        """
        #pylint: disable=no-member
        if group_distinguishedname.lower() in [dn.lower() for dn in self.memberof]:
        #pylint: enable=no-member
            return True
        else:
            return False


    def group_samaccountnames(self, base_dn):
        """For the current ADUser instance, determine which
        groups the user is a member of and convert the
        group DistinguishedNames to sAMAccountNames.
        The resulting list of groups may not be complete
        if explicit_membership_only was set to
        True when the object factory method (user() or users()) was
        called.

        :param str base_dn: The base DN to search within

        :return: A list of groups (sAMAccountNames) for which the
            current ADUser instance is a member, sAMAccountNames
        :rtype: list
        """
        #pylint: disable=no-member
        mappings = self.samaccountnames(base_dn, self.memberof)
        #pylint: enable=no-member
        groups = [samaccountname for samaccountname in mappings.values()]
        if not groups:
            logging.info("%s - unable to retrieve any groups for the current ADUser instance",
                         self.samaccountname)
        return groups




#! /usr/bin/env python
#
# Copyright (C) 2015 zulily, llc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""pudl setup"""

from setuptools import setup, find_packages

execfile('pudl/version.py')

setup(name='pudl',
      version=__version__,
      description="""pudl is an Active Directory client library and CLI""",
      author='zulily, llc',
      author_email='opensource@zulily.com',
      packages=find_packages(),
      url='https://github.com/zulily/pudl',
      license='Apache License, Version 2.0',
      entry_points={
          'console_scripts': [
              'pudl = pudl.scripts.cli:main'
          ]
      },
      install_requires=[
          'python-ldap',
          'pyyaml'
      ],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 2.7',
          'Topic :: System :: Systems Administration',
          'Topic :: System :: Systems Administration :: Authentication/Directory',
          'Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP',
      ],
)



# Copyright (C) 2015 zulily, llc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ad_object"""

import copy
import importlib
import logging


class ADObject(object):
    """A base class for AD objects."""

    def __init__(self, adq):
        """ADObject constructor"""
        # Setup logging, assumes a root logger already exists with handlers
        self.logger = logging.getLogger(__name__)

        self.adq = adq


    def to_dict(self):
        """Prepare a minimal dictionary with keys mapping to attributes for
        the current instance.

        """
        o_copy = copy.copy(self)
        # Remove some stuff that is not likely related to AD attributes
        for attribute in dir(self):
            if attribute == 'logger' or attribute == 'adq':
                try:
                    delattr(o_copy, attribute)
                except AttributeError:
                    pass

        return o_copy.__dict__


    def samaccountname(self, base_dn, distinguished_name):
        """Retrieve the sAMAccountName for a specific DistinguishedName

        :param str base_dn: The base DN to search within
        :param list distinguished_name: The base DN to search within
        :param list attributes: Object attributes to populate, defaults to all

        :return: A populated ADUser object
        :rtype: ADUser
        """
        mappings = self.samaccountnames(base_dn, [distinguished_name])

        try:
            # Usually we will find a match, but perhaps not always
            return mappings[distinguished_name]
        except KeyError:
            logging.info("%s - unable to retrieve object from AD by DistinguishedName",
                         distinguished_name)


    def samaccountnames(self, base_dn, distinguished_names):
        """Retrieve the sAMAccountNames for the specified DNs

        :param str base_dn: The base DN to search within
        :param list distinguished_name: A list of distinguished names for which to
            retrieve sAMAccountNames

        :return: Key/value pairs mapping DistinguishedName to sAMAccountName
        :rtype: dict
        """
        attributes = ['sAMAccountName']
        search_filter = '(|{0})'.format(''.join(['(DistinguishedName={0})'.format(dn)
                                                 for dn in distinguished_names]))
        logging.debug('%s Search filter: %s', self.__class__.__name__, search_filter)

        results = self.adq.search(base_dn, search_filter, attributes)

        mappings = {result[0]: result[1]['sAMAccountName'][0] for result in results}

        return mappings


    def _object_factory(self, search_result):
        """Given a single search result, create and return an object

        :param tuple search_result: a single search result returned by an LDAP query,
            position 0 is the DN and position 1 is a dictionary of key/value pairs

        :return: A single AD object instance
        :rtype: Object (ADUser, ADGroup, etc.)

        """
        class_name = self.__class__.__name__
        module = self.__module__
        logging.debug('Creating object of type %s for DN: %s', class_name, search_result[0])
        module = importlib.import_module('{0}'.format(module))
        class_ = getattr(module, class_name)
        ado = class_(self.adq)

        # A unique set of all attribute names found
        attribute_names = set()
        # A unique set
        multiples = set()
        for k in search_result[1].keys():
            if k not in attribute_names:
                attribute_names.add(k)
            else:
                multiples.add(k)
        for k, val in search_result[1].iteritems():
            if k in multiples and not hasattr(ado, k):
                setattr(ado, k.lower(), list())
            if hasattr(ado, k):
                value = getattr(ado, k)
                if len(val) == 1:
                    value.append(val[0])
                else:
                    value.append(val)
            else:
                if len(val) == 1:
                    setattr(ado, k.lower(), val[0])
                else:
                    setattr(ado, k.lower(), val)


        logging.debug('Attributes and values for %s object (DN: %s): %s', class_name,
                      search_result[0], ado.__dict__)

        return ado



# Copyright (C) 2015 zulily, llc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ad_computer"""

import logging

from pudl.ad_object import ADObject

class ADComputer(ADObject):
    """A class to represent AD computer objects.  Includes a number of
    helper methods, particularly object-factory related.

    ADComputer objects have minimal depth, with attributes set to
    strings or lists.  Available attributes are dependent
    on the results returned by the LDAP query.
    """


    def computer(self, base_dn, samaccountname, attributes=()):
        """Produces a single, populated ADComputer object through the object factory.
        Does not populate attributes for the caller instance.

        :param str base_dn: The base DN to search within
        :param str samaccountname: The computer's sAMAccountName
        :param list attributes: Object attributes to populate, defaults to all

        :return: A populated ADComputer object
        :rtype: ADComputer
        """

        computers = self.computers(base_dn, samaccountnames=[samaccountname], attributes=attributes)

        try:
            # Usually we will find a match, but perhaps not always
            return computers[0]
        except IndexError:
            logging.info("%s - unable to retrieve object from AD by sAMAccountName", samaccountname)


    def computers(self, base_dn, samaccountnames=(), attributes=()):
        """Gathers a list of ADComputer objects

        :param str base_dn: The base DN to search within
        :param list samaccountnames: A list of computer names for which objects will be
            created, defaults to all computers if unspecified
        :param list attributes: Object attributes to populate, defaults to all

        :return: A list of populated ADComputer objects
        :rtype: list
        """
        ad_computers = []

        search_filter = '(&(objectClass=computer){0})'
        # If no samaccountnames specified, filter will pull all computer objects under
        # base_dn
        if not samaccountnames:
            search_filter = search_filter.format('(sAMAccountName=*)')
        else:
            if len(samaccountnames) == 1:
                computer_names = '(sAMAccountName={0})'.format(samaccountnames[0])
            else:
                computer_names = '(|{0})'.format(''.join(['(sAMAccountName={0})'.\
                                                          format(computer) for computer
                                                          in samaccountnames]))

            search_filter = search_filter.format(computer_names)


        logging.debug('%s Search filter: %s', self.__class__.__name__, search_filter)

        results = self.adq.search(base_dn, search_filter, attributes)

        for search_result in results:
            adc = self._object_factory(search_result)
            ad_computers.append(adc)

        return ad_computers

